import logging
from typing import Any, Dict

from aiohttp import ClientSession

_LOGGER = logging.getLogger(__name__)


class API:
    _http_base_url: str
    _token: str

    def __init__(
        self,
        host: str,
        token: str,
        scheme: str = "https",
        port: int = 8443,
        version: str = "v1"
    ) -> None:
        self._http_base_url = f"{scheme}://{host}:{port}/{version}"
        self._token = token

    async def get(self, path: str) -> Any:
        url = f"{self._http_base_url}{path}"
        _LOGGER.debug("Making GET request to %s", url)
        client_session = ClientSession()
        try:
            async with client_session.get(
                url,
                headers=self._headers(),
                ssl=False,
                timeout=30
            ) as res:
                res.raise_for_status()
                return await res.json()
        finally:
            await client_session.close()

    async def patch(self, path: str, json: Dict) -> None:
        url = f"{self._http_base_url}{path}"
        _LOGGER.debug("Making GET request to %s", url)
        client_session = ClientSession()
        try:
            async with client_session.patch(
                url,
                json=json,
                headers=self._headers(),
                ssl=False,
                timeout=30
            ) as res:
                res.raise_for_status()
        finally:
            await client_session.close()

    def _headers(self) -> Dict[str, Any]:
        return {"Authorization": f"Bearer {self._token}"}



from typing import Dict, List

from aiodirigera.api import API
from aiodirigera.api_model import HubStatus, DeviceStatus


class Hub:
    _delegate: API = None

    def __init__(
        self,
        host: str,
        token: str,
        scheme: str = "https",
        port: int = 8443,
        version: str = "v1"
    ) -> None:
        self._delegate = API(
            host,
            token,
            scheme=scheme,
            port=port,
            version=version
        )

    async def get_hub_status(self) -> HubStatus:
        raw = await self._delegate.get("/hub/status")
        return HubStatus(**raw)

    async def get_devices(self) -> List[DeviceStatus]:
        raw = await self._delegate.get("/devices")
        return [DeviceStatus(**x) for x in raw]

    async def get_device(self, id: str) -> DeviceStatus:
        raw = await self._delegate.get(f"/devices/{id}")
        return DeviceStatus(**raw)

    async def update_device(self, id: str, data: Dict) -> None:
        await self._delegate.patch(f"/devices/{id}", data)



# Copied and modified from Leggin/dirigera under MIT License

import string
import hashlib
import random
import socket
import base64

from aiohttp import ClientSession

ALPHABET = f"_-~.{string.ascii_letters}{string.digits}"
CODE_LENGTH = 128


def random_char(alphabet: str) -> str:
    return alphabet[random.randrange(0, len(alphabet))]


def random_code(alphabet: str, length: int) -> str:
    return "".join([random_char(alphabet) for _ in range(0, length)])


def code_challenge(code_verifier: str) -> str:
    sha256_hash = hashlib.sha256()
    sha256_hash.update(code_verifier.encode())
    digest = sha256_hash.digest()
    sha256_hash_as_base64 = (
        base64.urlsafe_b64encode(digest).rstrip(b"=").decode("us-ascii")
    )
    return sha256_hash_as_base64


async def send_challenge(
    ip_address: str,
    code_verifier: str,
    client_session: ClientSession = ClientSession()
) -> str:
    url = f"https://{ip_address}:8443/v1/oauth/authorize"
    params = {
        "audience": "homesmart.local",
        "response_type": "code",
        "code_challenge": code_challenge(code_verifier),
        "code_challenge_method": "S256",
    }

    try:
        async with client_session.get(
            url,
            params=params,
            ssl=False,
            timeout=30
        ) as res:
            res.raise_for_status()
            res_json = await res.json()
            return res_json["code"]
    finally:
        await client_session.close()


async def get_token(
    ip_address: str,
    code: str,
    code_verifier: str,
    client_session: ClientSession = ClientSession()
) -> str:
    url = f"https://{ip_address}:8443/v1/oauth/token"
    data = str(
        "code="
        + code
        + "&name="
        + socket.gethostname()
        + "&grant_type="
        + "authorization_code"
        + "&code_verifier="
        + code_verifier
    )
    headers = {"Content-Type": "application/x-www-form-urlencoded"}

    try:
        async with client_session.post(
            url,
            data=data,
            headers=headers,
            ssl=False,
            timeout=30
        ) as res:
            res.raise_for_status()
            res_json = await res.json()
            return res_json["access_token"]
    finally:
        await client_session.close()



from typing import Optional, List

from dataclasses import dataclass

# Generic


@dataclass
class Capabilities:
    canSend: List[str]
    canReceive: List[str]


# Hub Specific


@dataclass
class UserConsents:
    name: str
    value: str


@dataclass
class HubAttributes:
    customName: str
    model: str
    manufacturer: str
    firmwareVersion: str
    hardwareVersion: str
    serialNumber: str
    identifyStarted: str
    identifyPeriod: str
    otaStatus: str
    otaState: str
    otaProgress: str
    otaPolicy: str
    otaScheduleStart: str
    otaScheduleEnd: str
    permittingJoin: str
    backendConnected: str
    backendConnectionPersistent: str
    backendOnboardingComplete: str
    backendRegion: str
    backendCountryCode: str
    userConsents: List[UserConsents]
    logLevel: int
    coredump: bool
    timezone: str
    nextSunSet: str
    nextSunRise: str
    homestateValue: str
    homestateLastChanged: str
    countryCode: str
    isOn: bool


@dataclass
class HubStatus:
    id: str
    relationId: str
    type: str
    deviceType: str
    createdAt: str
    isReachable: bool
    lastSeen: str
    attributes: HubAttributes
    capabilities: Capabilities
    deviceSet: List[str]
    remoteLinks: List[str]
    apiVersion: str


# Device Specific


@dataclass
class DeviceAttributes:
    customName: str
    model: str
    manufacturer: str
    firmwareVersion: str
    hardwareVersion: str
    serialNumber: str
    productCode: Optional[str] = None  # Only on Ikea devices
    isOn: Optional[bool] = None  # Only on On/Off devices
    startupOnOff: Optional[str] = None  # Only on On/Off devices
    lightLevel: Optional[bool] = None  # Only on lights
    currentTemperature: Optional[str] = None  # Only on environment sensors
    currentRH: Optional[str] = None  # Only on environment sensors
    identifyStarted: Optional[str] = None  # Only on Ikea devices
    identifyPeriod: Optional[int] = None  # Only on Ikea devices
    permittingJoin: Optional[bool] = None  # ???
    otaStatus: Optional[str] = None  # ???
    otaState: Optional[str] = None  # ???
    otaProgress: Optional[int] = None  # ???
    otaPolicy: Optional[str] = None  # ???
    otaScheduleStart: Optional[str] = None  # ???
    otaScheduleEnd: Optional[str] = None  # ???


@dataclass
class DeviceRoom:
    id: str
    name: str
    color: str
    icon: str


@dataclass
class DeviceStatus:
    id: str
    type: str
    deviceType: str
    createdAt: str
    isReachable: bool
    lastSeen: str
    attributes: DeviceAttributes
    capabilities: Capabilities
    room: DeviceRoom
    deviceSet: List[str]
    remoteLinks: List[str]
    isHidden: bool

    def __post_init__(self):
        self.attributes = DeviceAttributes(**self.attributes)
        self.capabilities = Capabilities(**self.capabilities)
        self.room = DeviceRoom(**self.room)
